/*
** Module   : ScrollableSelectionListGuiImpl.java
** Abstract : scrollable selection list implementation 
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150423 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Changing scrollbar painting approach.  Several minor UI fixes.  Adding method
**                  to help dismissing of the drop-down on outside mouse press.  Adding mouse
**                  hover tracking to change the current row of the drop-down list items.
** 004 EVL 20150720 Adding protection from getting negative index in draw() method.  Fix for issue
**                  with entry field editing after item select in simple mode of combo-box.  In
**                  this case isLeaveEventRequired() should return false.  Fix for TAB/BACK_TAB
**                  key processing issue.
** 005 EVL 20150814 Inherit row height from owning combo-box.
** 006 HC  20150831 Scrolling support overhaul.
** 007 CA  20150910 Added mouse actions.
** 008 CA  20150925 When the inner list is losing focus, the drop-down must be exited.
** 009 HC  20151013 Changes to extend GUI multi-window focus management and ACTIVE-WINDOW system
**                  handle processing to modal windows.
** 010 EVL 20151020 Fix for drop-down drawing issue with scrolling by scrollbar.
** 011 HC  20151029 Fixed drop-down scroll list width calculation, it must not be smaller than
**                  the parent's width.
**         20151103 Improved GUI COMBO-BOX layout.
**         20151110 GUI combo-box fixes.
**         20151118 Restored initial logic in leaveAndRepostMouse() removing exitDropDown() for
**                  combo-box body from the end of this method.  This exit cal is not required and
**                  can produce the regressions.  All neccessary drop-down exit cleanups are
**                  performing as reaction for ESC key artifitially generated by the
**                  leaveAndRepostMouse() method. This is event order sensitive method, we need
**                  to keep the right events ordering.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Scrollable selection list GUI implementation.  To be used to support combo box and selection
 * list widgets. 
 */
@SuppressWarnings("unchecked")
public class ScrollableSelectionListGuiImpl
extends ScrollableList<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Widget configuration */
   private BaseConfig config;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Row height */
   private int rowHeight;

   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected item text color. */
   private ColorRgb selectionFore;

   /** Disabled state text color. */
   private ColorRgb disabledFore;
   
   /** The curent top row in the list to draw. */
   private int rowTop;

   /**
    * Create component using given configuration and data model.
    * 
    * @param    box
    *           The instance with which this object must be associated.
    * @param    model
    *           The data model that represents the contents of the list.
    */
   ScrollableSelectionListGuiImpl(ComboBox<GuiOutputManager> box, ListModel<String> model)
   {
      super(box, model);
      // initialize screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // and font resolver
      gf = new GuiFontResolver(WindowManager.resolveWindow(this), gd, box.config(), false);
      // color initializing
      selectionBack = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      selectionFore = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      disabledFore  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      // top row number will be initialized later in setTopRow()
      
      // exit drop-down makes sense only for a drop-down child
      if (parent(DropDown.class) != null)
      {
         addFocusListener(new FocusListener()
         {
            @Override
            public void onFocusLost(FocusEvent event)
            {
               ScrollableSelectionListGuiImpl.this.box.exitDropDown();
            }
            
            @Override
            public void onFocusGained(FocusEvent event)
            {
               // no-op
            }
         });
      }
   }

   /**
    * Draws this component. Overrides draw() in JList.
    */
   @Override
   public void draw()
   {
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = box.config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_WINDOW, SysColor.COLOR_BTNTEXT);

      // determine colors to use
      gc.colors();

      // get actual rows to view
      final int rowsVisible = visibleRows();
      
      // resolve font
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      // TODO: remove this line and rely on the set size, modifying
      //       the size here is very confusing and may lead
      //       to unexpected results
      d.height = rowsVisible * rowHeight;

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, d);

      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);
            
            // fill the background
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);

            // draw entry field text
            gd.setGuiFont(fntCache.font);
            gd.setFontStyle(fntCache.font.style);
            
            int x = po.x;
            int y = po.y;
            int fontSize = fntCache.font.pointSize;
            int modelSize = model().size();
            // adjust top row number to have last item from the list as bottom number 
            if ((rowTop + rowsVisible) > modelSize)
            {
               // protect from getting negative index
               rowTop = modelSize > rowsVisible ? modelSize - rowsVisible : 0;
            }
            
            // check if the list is enabled
            boolean enabled = isEnabled();
            
            // draw items from the top to bottom
            for (int i = rowTop; i < modelSize && i < rowsVisible + rowTop;
                            i++, y+= rowHeight)
            {
               if (y < 0)
                  continue;
               
               // draw next visible item in the list          
               drawItem(i, x, y, d.width, rowHeight, fontSize, enabled);
            }
         }
      });
   }
   
   @Override
   public void scroll(Integer newXPosition, Integer newYPosition)
   {
      rowTop = newYPosition;
       
      // and repaint the list
      repaint();
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(this, e);
            
            // calculate clicked line
            int row = n.y / rowHeight + rowTop;

            // in combo-box it is possible to generate selection for already selected item
            // go to selected item
            super.ensureRowVisible(row, false);
            // select new item
            selectItem();
            // and generate LEAVE to dismiss drop-down in respective mode
            if (getBox().config().mode != ComboBoxConfig.Mode.SIMPLE)
            {
               KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_LEAVE);
               EventManager.postEvent(ki);
            }
            
            repaint(); 
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Notification of a mouse moving event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      if (!isEnabled() || getBox().config().mode == ComboBoxConfig.Mode.SIMPLE)
      {
         return;
      }
      
      int rowHeight = ((ComboBoxGuiImpl)box).getRowHeight();
      
      synchronized (this)
      {
         if (model().size() > 0)
         {
            NativePoint n = translate(this, e);

            // calculate line under the mouse
            int row = n.y / rowHeight + rowTop;

            // restrict repainting only when we are over the new line
            if (row != currentRow())
            {
               // do not select new item, only make it the current one
               super.ensureRowVisible(row, false);
               
               // no need to call additional repaint here, will be done in other place
            }
         }
      }
   }
   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI mode 
   }
   
   /**
    * Intercept key events and handle some special cases which are not
    * handled by base class.
    * 
    * @param    kvt
    *           Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput kvt)
   {
      ComboBoxGuiImpl cbbRef = (ComboBoxGuiImpl) getBox();
      int key = kvt.actionCode();

      // disable ENTER key itself as event generator to avoid value change duplication
      // this event has already registered as drop-down exit trigger
      if (key == Keyboard.KA_RETURN)
      {
         kvt.consume();
      }
      // in simple mode TAB/BACK_TAB keys should move the focus to next/prev widget
      // comparing to the combo-box this list belongs to
      else if ((key == Keyboard.KA_TAB || key == Keyboard.KA_BACK_TAB) &&
               cbbRef != null && cbbRef.config().mode == ComboBoxConfig.Mode.SIMPLE)
      {
         cbbRef.processFocusMoveKey(kvt);
      }
      else
      {
         super.processKeyEvent(kvt);
      }
   }

   /**
    * Check if the widget needs LEAVE event on special mouse press conditions.
    * 
    * @param    eventSrc
    *           The event source.
    * @param    evtToTest
    *           The mouse event to check.  Must be press, release or click.
    * 
    * @return   <code>TRUE</code> if a mouse action should cause the <code>LEAVE</code> event
    *           to be generated.
    */
   @Override
   public boolean isLeaveEventRequired(Widget<GuiOutputManager> eventSrc, MouseEvt evtToTest)
   {
      // the event source must be not null
      if (eventSrc != null)
      {
         // check for possible clicks inside drop-down but not in the list itself
         // it can be the drop-down buttons for example
         Widget<GuiOutputManager> wAncest = eventSrc.ancestor();
         
         // if we are going outside the drop-down - send leave to it,
         // the conditions to be met:
         // - the widget is not a part of the simple mode combo-box
         // - we are not in other parts of this drop-down
         // - we are not in selection list or parts of the combo-box owning this selecton list
         // - the event is not artificially reposted
         if (getBox().config().mode != ComboBoxConfig.Mode.SIMPLE &&
             (wAncest == null || !(wAncest instanceof DropDownGuiImpl)) &&
             eventSrc != this && eventSrc != this.getBox() && evtToTest.isOriginal())
         {
            return true;
         }
      }

      return false;
   }
   
   /**
    * Generates leave event as reaction on mouse event.  The mouse event is reposted to guarantee
    * the proper event sequence.
    * 
    * @param    evtToRepost
    *           The mouse event to modify and repost.
    */
   @Override
   public void leaveAndRepostMouse(MouseEvt evtToRepost)
   {
      // only press has special meaning to insert leave
      if (evtToRepost.getEvent().getID() == MouseEvt.MOUSE_PRESSED)
      {
         // in GUI leave event causes the drop-down to dismiss
         // do NOT call box.exitDropDown() in this method it must be done automatically as
         // reaction for ESCAPE key
         KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Key.VK_ESCAPE);
         EventManager.postEvent(ki);
      }
      // mark the event as artificially reposted
      evtToRepost.setOriginal(false);
      // repost original event
      EventManager.postEvent(evtToRepost);
   }

   /**
    * Determine the mouse actions processed by this widget.
    * 
    * @return   See above.
    */
   @Override
   protected int[] mouseActions()
   {
      return new int[] { MouseEvent.MOUSE_CLICKED, MouseEvent.MOUSE_MOVED };
   }
   
   /**
    * Set item number to be displayed on the top of the drop-down list.
    * 
    * @param   index
    *          The index of the top item.
    */
   @Override
   protected void setTopRow(int index)
   {
      ensureRowVisible(index, false);
      rowTop = fixTopRow(index);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * Select row with specified index.
    * 
    * @param   index
    *          Item index.
    */
   @Override
   public void select(int index)
   {
      super.select(index);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * Select specified range of items.
    * 
    * @param   from
    *          Start index in the range.
    * @param   to
    *          End index in the range.
    */
   @Override
   public void select(int from, int to)
   {
      super.select(from, to);
      postScrollEvent(null, rowTop);
   }
   
   /**
    * The scrollable list's boundary does not keep track of the view size - it posts a repaint 
    * event for the entire list, regardless how small a part is visible.
    * <p>
    * If the list is attached to a parent, then the repaint will be posted for its parent.
    */
   @Override
   public void repaint()
   {
      if (parent() != null)
      {
         parent().repaint();
      }
   }
   
   /**
    * Shifts current row by specified delta.
    * 
    * @param   delta
    *          the number of list items to shift current row by
    * @param   scroll
    *          indicates whether list should be scrolled
    *          to the shift row
    */
   @Override
   protected void shiftCurrentRow(int delta, boolean scroll)
   {
      super.shiftCurrentRow(delta, false);
  
      int origRowTop = rowTop;
      
      // recalc top row changes
      if (currentRow() < rowTop)
      {
         // upper than visible region
         rowTop = currentRow(); 
      }
      else if (currentRow() >= rowTop + visibleRows())
      {
         // lower than visible region
         rowTop = currentRow() - visibleRows() + 1;
      }
      
      // and notify main combo-box about new selection for keyboard event origin case
      if (scroll)
      {
         selectItem();
      }
   }
   
   /**
    * Calculate the scroll Y position based on the current state. 
    * 
    * @return   the new scroll Y position
    */
   @Override
   protected int calculateScrollYPosition()
   {
      return fixTopRow(rowTop);
   }

   /**
    * Draws the next text item in the list inside visible part of the drop-down.
    * 
    * @param    number
    *           The number of the item to draw.
    * @param    x
    *           The X coordinate of the enclosing rectangle to draw item.
    * @param    y
    *           The Y coordinate of the enclosing rectangle to draw item.
    * @param    rw
    *           The pixel width of the enclosing rectangle to draw item.
    * @param    rh
    *           The pixel height of the enclosing rectangle to draw item.
    * @param    fontSize
    *           The vertical size of font to draw item text.
    * @param    enabled
    *           Flag indicating the enabled state of the list.
    */
   private void drawItem(int number, int x, int y, int rw, int rh, int fontSize, boolean enabled)
   {
      if (number == currentRow())
      {
         gd.setColor(selectionBack);
         gd.fillRect(x, y, rw, rh);
      }

      if (enabled && number == currentRow())
      {
         gd.setColor(selectionFore);
      }
      else if (enabled)
      {
         gd.setColor(gc.fgColor);
      }
      else
      {
         gd.setColor(disabledFore);
      }

      gd.drawString(getItemText(number), x + ComboBoxGuiImpl.TEXT_SHIFT, y + (rh + fontSize)/2);
   }
   
   /**
    * Converts the passed in value to a valid row unless the
    * value is already valid.
    * 
    * @param    desiredTopRow
    *           The value to check and fix.
    *          
    * @return   valid row value
    */
   private int fixTopRow(int desiredTopRow)
   {
      if (desiredTopRow < 0)
      {
         return 0;
      }
      
      int maxTopRow = model().size() - visibleRows() - 1;
      
      if (maxTopRow < 0)
      {
         maxTopRow = 0;
      }
      
      if (desiredTopRow > maxTopRow)
      {
         return maxTopRow;
      }
      
      return desiredTopRow;
   }
}
